@@ -49,24 +49,25 @@ impl Serialize for EventValues {
     }
 }
 
-/// The content of an Event, sorted by message types
-#[derive(Clone,Debug,Default)]
-pub struct EventBody {
-    values: IndexMap<MessageType,EventValues>,
-}
 
 #[derive(Clone,Debug)]
 pub struct Event {
     node: Option<Vec<u8>>,
     id: EventID,
-    body: EventBody,
+    body: IndexMap<MessageType,EventValues>,
+}
+
+impl Event {
+    fn new(node: Option<Vec<u8>>, id: EventID) -> Self {
+        Event { node, id, body: IndexMap::with_capacity(5) }
+    }
 }
 
 impl Serialize for Event {
     fn serialize<S>(&self, s: S) -> Result<S::Ok, S::Error>
     where S: Serializer
     {
-        let mut length = 1 + self.body.values.len();
+        let mut length = 1 + self.body.len();
         if let Some(_) = self.node {
             length += 1
         };
@@ -77,7 +78,7 @@ impl Serialize for Event {
             map.serialize_key("NODE")?;
             map.serialize_value(&node.as_slice().to_quoted_string())?;
         }
-        for (k,v) in &self.body.values {
+        for (k,v) in &self.body {
             map.serialize_entry(&k,&v)?;
         }
         map.end()
@@ -87,7 +88,7 @@ impl Serialize for Event {
 /// Coalesce collects Audit Records from individual lines and assembles them to Events
 pub struct Coalesce<'a> {
     /// Events that are being collected/processed
-    inflight: BTreeMap<(Option<Vec<u8>>, EventID), EventBody>,
+    inflight: BTreeMap<(Option<Vec<u8>>, EventID), Event>,
     /// Event IDs that have been recently processed
     done: HashSet<(Option<Vec<u8>>, EventID)>,
     /// Timestamp for next cleanup
@@ -103,6 +104,9 @@ pub struct Coalesce<'a> {
     pub execve_argv_string: bool,
     pub execve_env: HashSet<Vec<u8>>,
 
+    pub proc_label_keys: HashSet<Vec<u8>>,
+    pub proc_propagate_labels: HashSet<Vec<u8>>,
+
     pub translate_universal: bool,
     pub translate_userdb: bool,
 }
@@ -210,6 +214,8 @@ impl<'a> Coalesce<'a> {
             execve_argv_list: true,
             execve_argv_string: false,
             execve_env: HashSet::new(),
+            proc_label_keys: HashSet::new(),
+            proc_propagate_labels: HashSet::new(),
             translate_universal: false,
             translate_userdb: false,
         }
@@ -229,8 +235,8 @@ impl<'a> Coalesce<'a> {
             .cloned()
             .collect::<Vec<_>>();
         for (node, id) in node_ids {
-            if let Some(body) = self.inflight.remove(&(node.clone(), id)) {
-                self.emit_event(&mut Event{node, id, body});
+            if let Some(event) = self.inflight.remove(&(node.clone(), id)) {
+                self.emit_event(event);
             }
         }
     }
@@ -273,16 +279,23 @@ impl<'a> Coalesce<'a> {
         return None;
     }
 
-    /// Rewrite EventBody to normal form
+    /// Rewrite event to normal form
     ///
     /// This function
     /// - turns SYSCALL/a* fields into a single an ARGV list
     /// - turns EXECVE/a* and EXECVE/a*[*] fields into an ARGV list
     /// - turns PROCTITLE/proctitle into a (abbreviated) ARGV list
     /// - translates *uid, *gid, syscall, arch, sockaddr if configured to do so.
-    fn normalize_eventbody(&mut self, eb: &mut EventBody) {
-        for (typ,ev) in eb.values.iter_mut() {
-            match (typ,ev) {
+    /// - collects environment variables for EXECVE
+    /// - registers process in shadow process table for EXECVE
+    fn transform_event(&mut self, ev: &mut Event) {
+        let mut pid: Option<u32> = None;
+        let mut ppid: Option<u32> = None;
+        let mut comm: Option<Vec<u8>> = None;
+        let mut exe: Option<Vec<u8>> = None;
+        let mut key: Option<Vec<u8>> = None;
+        for tv in ev.body.iter_mut() {
+            match tv {
                 (&SYSCALL, EventValues::Single(rv)) => {
                     rv.raw.reserve(
                         if self.translate_universal { 16 } else { 0 } +
@@ -293,29 +306,46 @@ impl<'a> Coalesce<'a> {
                     let mut arch = None;
                     let mut syscall = None;
                     for (k,v) in &rv.elems.clone() {
-                        match k {
-                            Key::Arg(_,None) => {
+                        match (k,v) {
+                            (Key::Arg(_,None), _) => {
                                 // FIXME: check argv length
                                 argv.push(v.clone());
                                 continue;
                             },
-                            Key::ArgLen(_) => continue,
-                            Key::Name(r) => {
+                            (Key::ArgLen(_), _) => continue,
+                            (Key::Name(r), Value::Number(n)) => {
+                                let name = &rv.raw[r.clone()];
+                                match (name, n) {
+                                    (b"arch", Number::Hex(n)) if self.translate_universal =>
+                                        arch = Some((r.clone(), n.clone())),
+                                    (b"syscall", Number::Dec(n)) if self.translate_universal =>
+                                        syscall = Some((r.clone(), n.clone())),
+                                    (b"pid", Number::Dec(n)) => pid = Some(*n as u32),
+                                    (b"ppid", Number::Dec(n)) => ppid = Some(*n as u32),
+                                    _ => (),
+                                }
+                            },
+                            (Key::Name(r), v) => {
                                 let name = &rv.raw[r.clone()];
                                 match name {
-                                    b"arch" if self.translate_universal => {
-                                        if let Value::Number(Number::Hex(n)) = v {
-                                            arch = Some((r.clone(), n.clone()));
-                                        }
-                                    },
-                                    b"syscall" if self.translate_universal => {
-                                        if let Value::Number(Number::Dec(n)) = v {
-                                            syscall = Some((r.clone(), n.clone()));
-                                        }
-                                    },
-                                    b"ARCH"|b"SYSCALL" if self.translate_universal => continue,
-                                    _ => ()
-                                }
+                                    b"ARCH" =>
+                                        if let &Some(_) = &arch { continue },
+                                    b"SYSCALL" =>
+                                        if let &Some(_) = &syscall { continue },
+                                    b"key" =>
+                                        if let Value::Str(r, _) = v {
+                                            key = Some(rv.raw[r.clone()].into());
+                                        },
+                                    b"comm" =>
+                                        if let Value::Str(r,_) = v {
+                                            comm = Some(rv.raw[r.clone()].into());
+                                        },
+                                    b"exe" =>
+                                        if let Value::Str(r,_) = v {
+                                            exe = Some(rv.raw[r.clone()].into());
+                                        },
+                                    _ => (),
+                                };
                             },
                             _ => if let Some((k,v)) = self.translate_userdb(rv, k, v) {
                                 translated.push_back((k,v));
@@ -373,13 +403,47 @@ impl<'a> Coalesce<'a> {
                             _ => new.push((k.key.clone(), v.value.clone())),
                         };
                     }
+
+                    // ARGV
                     if self.execve_argv_list {
                         new.push((Key::Literal("ARGV"), Value::List(argv.clone())));
                     }
+                    // ARGV_STR
                     if self.execve_argv_string {
                         new.push((Key::Literal("ARGV_STR"), Value::StringifiedList(argv.clone())));
                     }
+                    // ENV
+                    match pid {
+                        Some(pid) if !self.execve_env.is_empty() => {
+                            if let Ok(vars) = get_environ(pid, |k| self.execve_env.contains(k) ) {
+                                let map = vars.iter()
+                                    .map(|(k,v)| (rv.put(k), rv.put(v)))
+                                    .collect();
+                                new.push( (Key::Literal("ENV"), Value::Map(map)) );
+                            }
+                        },
+                        _ => (),
+                    };
+
                     rv.elems = new;
+
+                    // register process, add propagated labels from
+                    // parent if applicable
+                    if let (Some(pid), Some(ppid)) = (pid, ppid) {
+                        let argv = argv.iter().filter_map(
+                            |v| match v {
+                                Value::Str(r,_) => Some(Vec::from(&rv.raw[r.clone()])),
+                                _ => None
+                            }
+                        ).collect();
+                        self.processes.add_process(pid, ppid, ev.id, comm.clone(), exe.clone(), argv);
+
+                        if let Some(parent) = self.processes.get_process(ppid) {
+                            for l in self.proc_propagate_labels.intersection(&parent.labels) {
+                                self.processes.add_label(pid, &l);
+                            }
+                        }
+                    }
                 },
                 (&SOCKADDR, EventValues::Multi(rvs)) => {
                     for mut rv in rvs {
@@ -444,49 +508,66 @@ impl<'a> Coalesce<'a> {
                 },
             }
         }
-    }
 
-    /// Add environment variables to event body
-    fn augment_execve_env(&mut self, execve: &mut Record, pid: u64) {
-        if self.execve_env.is_empty() {
-            return;
+        if let (Some(pid), Some(key)) = (&pid, &key) {
+            if self.proc_label_keys.contains(key) {
+                self.processes.add_label(*pid, key);
+            }
         }
-        if let Ok(vars) = get_environ(pid, |k| self.execve_env.contains(k) ) {
-            let map = vars.iter()
-                .map(|(k,v)| (execve.put(k), execve.put(v)))
-                .collect();
-            execve.elems.push( (Key::Literal("ENV"), Value::Map(map)) );
+
+        // PARENT_INFO
+        if let Some(ppid) = ppid {
+            if let Some(p) = self.processes.get_process(ppid) {
+                let mut pi = Record::default();
+                if let Some(id) = p.event_id {
+                    let r = pi.put(&format!("{}", id).as_bytes());
+                    pi.elems.push((Key::Literal("ID"), Value::Str(r, Quote::None)));
+                }
+                if let Some(comm) = p.comm {
+                    let r = pi.put(&comm);
+                    pi.elems.push((Key::Literal("comm"), Value::Str(r, Quote::None)));
+                }
+                if let Some(exe) = p.exe {
+                    let r = pi.put(&exe);
+                    pi.elems.push((Key::Literal("exe"), Value::Str(r, Quote::None)));
+                }
+                let argv = p.argv.iter()
+                    .map(|v| Value::Str(pi.put(v), Quote::None))
+                    .collect::<Vec<_>>();
+                if self.execve_argv_list {
+                    pi.elems.push((Key::Literal("ARGV"), Value::List(argv.clone())));
+                }
+                if self.execve_argv_string {
+                    pi.elems.push((Key::Literal("ARGV_STR"), Value::StringifiedList(argv.clone())));
+                }
+                let kv = (Key::Name(pi.put(b"ppid")), Value::Number(Number::Dec(p.ppid as u64)));
+                pi.elems.push(kv);
+                ev.body.insert(PARENT_INFO, EventValues::Single(pi));
+            }
+        }
+        if let (Some(pid), Some(EventValues::Single(sc))) =
+            (pid, ev.body.get_mut(&SYSCALL))
+        {
+            let labels = self.processes.get_process(pid)
+                .and_then(|p| Some(p.labels))
+                .unwrap_or(HashSet::new());
+            if labels.len() > 0 {
+                let labels = labels.iter()
+                    .map(|l| Value::Str(sc.put(l), Quote::None))
+                    .collect::<Vec<_>>();
+                sc.elems.push((Key::Literal("LABELS"), Value::List(labels)));
+            }
         }
     }
 
     /// Do bookkeeping on event, transform, emit it via the provided
     /// output function.
-    fn emit_event(&mut self, e: &mut Event) {
-        self.done.insert((e.node.clone(), e.id));
+    fn emit_event(&mut self, mut ev: Event) {
+        self.done.insert((ev.node.clone(), ev.id));
 
-        self.normalize_eventbody(&mut e.body);
+        self.transform_event(&mut ev);
 
-        let mut pid = None;
-        if let Some(EventValues::Single(ref syscall)) = e.body.values.get(&SYSCALL) {
-            #[allow(unused_must_use)]
-            if let Some(EventValues::Single(ref execve)) = e.body.values.get(&EXECVE) {
-                self.processes.add_execve(&e.id, &syscall, &execve);
-            }
-            if let Some(v) = syscall.get(b"pid") {
-                if let Value::Number(Number::Dec(p)) = v.value {
-                    pid = Some(*p);
-                }
-            }
-        }
-
-        match (pid, e.body.values.get_mut(&EXECVE)) {
-            (Some(pid), Some(EventValues::Single(ref mut execve))) =>  {
-                self.augment_execve_env(execve, pid);
-            }
-            _ => (),
-        };
-
-        (self.emit_fn)(&e)
+        (self.emit_fn)(&ev)
     }
 
     /// Ingest a log line and add it to the coalesce object.
@@ -520,31 +601,27 @@ impl<'a> Coalesce<'a> {
             if self.done.contains(&nid) {
                 return Err(format!("duplicate event id {}", id).into());
             }
-            let body = self.inflight.remove(&nid)
+            let ev = self.inflight.remove(&nid)
                 .ok_or(format!("Event id {} for EOE marker not found", &id))?;
-            self.emit_event(&mut Event{node, id, body});
+            self.emit_event(ev);
         } else if typ.is_multipart() {
             // kernel-level messages
             if !self.inflight.contains_key(&nid) {
-                self.inflight.insert(nid.clone(), EventBody::default());
+                self.inflight.insert(nid.clone(), Event::new(node.clone(), id));
             }
-            let body = self.inflight.get_mut(&nid).unwrap();
-            match body.values.get_mut(&typ) {
+            let ev = self.inflight.get_mut(&nid).unwrap();
+            match ev.body.get_mut(&typ) {
                 Some(EventValues::Single(v)) => v.extend(rv),
                 Some(EventValues::Multi(v)) => v.push(rv),
                 None => match typ {
                     SYSCALL => {
-                        let pi = get_parent_info(&mut self.processes, &rv, self.execve_argv_list, self.execve_argv_string);
-                        body.values.insert(typ, EventValues::Single(rv));
-                        if let Some(pi) = pi {
-                            body.values.insert(PARENT_INFO, EventValues::Single(pi));
-                        }
+                        ev.body.insert(typ, EventValues::Single(rv));
                     },
                     EXECVE | PROCTITLE | CWD => {
-                        body.values.insert(typ, EventValues::Single(rv));
+                        ev.body.insert(typ, EventValues::Single(rv));
                     },
                     _ => {
-                        body.values.insert(typ, EventValues::Multi(vec!(rv)));
+                        ev.body.insert(typ, EventValues::Multi(vec!(rv)));
                     },
                 }
             };
@@ -553,39 +630,14 @@ impl<'a> Coalesce<'a> {
             if self.done.contains(&nid) {
                 return Err(format!("duplicate event id {}", id).into());
             }
-            let mut values = IndexMap::new();
-            values.insert(typ, EventValues::Single(rv));
-            self.emit_event(&mut Event{node, id, body: EventBody{values}});
+            let mut ev = Event::new(node, id);
+            ev.body.insert(typ, EventValues::Single(rv));
+            self.emit_event(ev);
         }
         Ok(())
     }
 }
 
-fn get_parent_info(pt: &mut ProcTable, rv: &Record, do_list: bool, do_string: bool) -> Option<Record> {
-    if let Some(v) = rv.get(b"ppid") {
-        if let Value::Number(Number::Dec(ppid)) = v.value {
-            if let Some(p) = pt.get_process(*ppid) {
-                let mut pi = Record::default();
-                let vs = p.argv.iter()
-                    .map(|v| Value::Str(pi.put(v), Quote::None))
-                    .collect::<Vec<_>>();
-                if do_string {
-                    let k = Key::Name(pi.put(b"ARGV_STR"));
-                    pi.elems.push((k, Value::StringifiedList(vs.clone())));
-                }
-                if do_list {
-                    let k = Key::Name(pi.put(b"ARGV"));
-                    pi.elems.push((k, Value::List(vs)));
-                }
-                let kv = (Key::Name(pi.put(b"ppid")), Value::Number(Number::Dec(p.ppid)));
-                pi.elems.push(kv);
-                return Some(pi);
-            }
-        }
-    }
-    return None;
-}
-
 impl Drop for Coalesce<'_> {
     fn drop(&mut self) { self.expire_inflight(u64::MAX) }
 }
