@@ -372,6 +372,10 @@ fn parse_unspec_value<'a>(input: &'a[u8], ty: MessageType, name: &[u8]) -> IResu
             delimited(tag("'"), take_while(|c| c != b'\''), tag("'")),
             |s| -> Result<_,()> { Ok(PValue::Str(s, Quote::Single)) }
         ),
+        map_res(
+            delimited(tag("\""), take_while(|c| c != b'"'), tag("\"")),
+            |s| -> Result<_,()> { Ok(PValue::Str(s, Quote::Double)) }
+        ),
         map_res(
             delimited(tag("{"), take_while(|c| c != b'}'), tag("}")),
             |s| -> Result<_,()> { Ok(PValue::Str(s, Quote::Braces)) }
@@ -705,6 +709,8 @@ mod test {
 
         let (_n, _t, _id, _rv) = parse(Vec::from(include_bytes!("testdata/line-netfilter.txt").as_ref()))?;
 
+        let (_n, _t, _id, _rv) = parse(Vec::from(include_bytes!("testdata/line-anom-abend.txt").as_ref()))?;
+
         Ok(())
     }
 }
